//==============================================================================
// FastFileDialog
//------------------------------------------------------------------------------
// Presents a simplified and much speeded up file dialog. 
// Apparently, the built-in file dialog tests every name in the 
// directory to see if it is a file or a directory, and builds the 
// file and directory lists based on this information.  Some of our
// trace directories contain three thousand files so this is prohibitively 
// slow.  Worse, the built-in file dialog does not save the file list
// between .show() calls, so you have to take this extrodinary hit 
// each and every time you open the dialog.   This dialog is several 
// times faster than the built-in one, and you only have to take the 
// hit once per directory.  It's still pretty slow, with 3000 files to 
// list, but a big improvement.  
//
// I believe the interface is the same, or at least a subset of, the 
// built-in file dialog, but I haven't explicitly verified this. 
//
//==============================================================================
//
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//       1         2         3         4         5         6         7         8

import java.awt.*;
import java.io.*;

//------------------------------------------------------------------------------
// * FastFileDialog
//------------------------------------------------------------------------------
//
public class FastFileDialog extends Dialog {

List			mFileList;
File			mDirect;
TextField		mTextField;
FilenameFilter	mFilter=null;
Panel			mButtonPanel;
Button			mbtnLoad,mbtnCancel,mbtnUpdate;
boolean         mLoad = false;

String			mDirName;
String			mFileName;

//------------------------------------------------------
// * FastFileDialog
//------------------------------------------------------
//
public FastFileDialog(Frame parent, String title) {
	super(parent, title, true);
	
	this.resize(250,400);
	this.setLayout(new BorderLayout());
	
	mTextField= new TextField("./",30);
	this.add("North",mTextField);
	
	mFileList = new List(10,false);
	mDirect = new File(mTextField.getText());
	String FilesInDir[] = mDirect.list(mFilter);
	for (int i =0;i<FilesInDir.length;i++){
		mFileList.addItem(FilesInDir[i]);
	}
	this.add("Center",mFileList);
	
	setupButtonPanel();
	this.add("South",mButtonPanel);
}

//------------------------------------------------------
// * setupButtonPanel
//------------------------------------------------------
//
protected void setupButtonPanel(){
	
	mButtonPanel = new Panel();
	
	mbtnLoad = new Button("Open");
	mbtnCancel = new Button("Cancel");
	mbtnUpdate = new Button("Update");
	
	mButtonPanel.add(mbtnLoad);
	mButtonPanel.add(mbtnUpdate);
	mButtonPanel.add(mbtnCancel);
	
}

//------------------------------------------------------
// * setFilter
//------------------------------------------------------
// An optional filter can be used to limit the files that
// are shown.   
//
public void setFilter(FilenameFilter theFilter){
	mFilter = theFilter;
}

//------------------------------------------------------
// * setDirectoryWithUpdate
//------------------------------------------------------
// Sets the directory to pathName, updating the list of 
// files displayed.  
//
public void setDirectoryWithUpdate(String pathName){
	if (pathName == ""){
		pathName=pathName+".";
	}

	// Add a slash if necessary and update text box. 
	if (!pathName.endsWith("/")){
		pathName = pathName+"/";
	}
	
	mTextField.setText(pathName);
	
	// Get the directory list
	mDirect = new File(pathName);
	String FilesInDir[] = mDirect.list(mFilter);
	
	mFileList.clear();
	
	// Add a list item for the parent directory
	mFileList.addItem("/..");
	
	// and then list the contents of the directory
	for (int i =0;i<FilesInDir.length;i++){
		mFileList.addItem(FilesInDir[i]);
	}
}

//------------------------------------------------------
// * setTextFieldToDirectory
//------------------------------------------------------
// Sets the directory to pathName, updating the list of 
// files displayed.  
//
public void setTextFieldToDirectory(String pathName){
	// Add a slash if necessary and update text box. 
	if (!pathName.endsWith("/")){
		pathName = pathName+"/";
	}
	mTextField.setText(pathName);
	
}

//------------------------------------------------------
// * GetDirNameFromTextField()
//------------------------------------------------------
//
public String GetDirNameFromTextField()
{
	String dirName = mTextField.getText();
	if (dirName.endsWith("/")){
		return(dirName);
	}
	else{
		dirName = dirName+"/";
		mTextField.setText(dirName);
		mTextField.repaint();
		return(dirName);
	}
}

//------------------------------------------------------
// * SetDirFromTextField()
//------------------------------------------------------
//
public void SetDirNameFromTextField()
{
	mDirName = GetDirNameFromTextField();
}

//------------------------------------------------------
// * SetFileAndDirFromTextField()
//------------------------------------------------------
//
public void SetFileAndDirFromTextField()
{
	String pathName = mTextField.getText();
	int lastSlash = pathName.lastIndexOf("/");
	mFileName = pathName.substring(lastSlash+1);
	mDirName = pathName.substring(0,lastSlash+1);
	setTextFieldToDirectory(mDirName);
	System.out.println("f="+mFileName+" d="+mDirName);
}

//------------------------------------------------------
// * SetFileFromSelection
//------------------------------------------------------
//
public void SetFileFromSelection()
{
	mFileName = mFileList.getSelectedItem();
}

//------------------------------------------------------
// * SetDirFromSelection
//------------------------------------------------------
//
public void SetDirNameFromSelection()
{
	String path = CreatePathFromTextAndSelection();
	mDirName = path;
}

//------------------------------------------------------
// * getFile
//------------------------------------------------------
//
public String getFile(){
	return(mFileName);
}

//------------------------------------------------------
// * getDirectory
//------------------------------------------------------
// Returns pathname in the directory window, appending 
// a trailing slash on it if necessary.  
//
public String getDirectory(){
	return(mDirName);
}

//------------------------------------------------------
// * load
//------------------------------------------------------
//
public boolean load(){
	return(mLoad);
}

//------------------------------------------------------
// * CreatePathFromTextAndSelection 
//------------------------------------------------------
// Combines the contents of the directory field with 
// the current selection, for the purpose of determining
// whether the current selection is a directory or a 
// file. 
//
//
public String CreatePathFromTextAndSelection()
{		
	String ParentPath;
	String testFile = mFileList.getSelectedItem();
	String testDir = GetDirNameFromTextField();
	
	// If the path is the parent directory extract the parent
	// dir name from the current path and return that...
	if (testFile == "/.."){
		int lastSlash = testDir.lastIndexOf("/");
		if (lastSlash > 0 ){
			ParentPath = testDir.substring(0,lastSlash-1);
			lastSlash = ParentPath.lastIndexOf("/");
			// If there was a second to last slash, make parent path
			// the path without the trailing dir name, otherwise we 
			// must be at the root level already.
			if (lastSlash > 0 ){
				ParentPath = ParentPath.substring(0,lastSlash);
			}
			else{
				ParentPath = "./";
			}
		}
		else{
			ParentPath = "./";
		}
		return(ParentPath);
	}
	else{
		return(testDir+testFile);
	}	
}


//------------------------------------------------------
// * SelectionIsDirectory
//------------------------------------------------------
//
public boolean SelectionIsDirectory()
{
	String testPath = CreatePathFromTextAndSelection();
	File testFile = new File(testPath);
	if (testFile.isDirectory()){
		return(true);
	}else{
		return(false);
	}
}

//------------------------------------------------------
// * SelectionIsFile
//------------------------------------------------------
//
public boolean SelectionIsFile()
{
	String testPath = CreatePathFromTextAndSelection();
	File testFile = new File(testPath);
	if (testFile.isFile()){
		return(true);
	}else{
		return(false);
	}
}

//------------------------------------------------------
// * TextFieldIsFile
//------------------------------------------------------
//
public boolean TextFieldIsFile(){
	String testName = mTextField.getText();
	File testFile = new File(testName);
	if (testFile.isFile()){
		return(true);
	}else{
		// Give it another shot with extension:
		testName=testName+".scf";
		testFile = new File(testName);
		if (testFile.isFile()){
			mTextField.setText(testName);
			return(true);
		}
		else{
			return(false);
		}
	}
}

//------------------------------------------------------
// * TextFieldIsDirectory
//------------------------------------------------------
//
public boolean TextFieldIsDirectory(){
	String testName = mTextField.getText();
	File testFile = new File(testName);
	if (testFile.isDirectory()){
		return(true);
	}else{
		return(false);
	}
}

//------------------------------------------------------
// * action
//------------------------------------------------------
//
public boolean action(Event event,Object arg){
	Object target = event.target;
	
	//System.out.println("id="+event.id);
	
	//----------------
	// Double Click
	//----------------
	if (event.id == Event.ACTION_EVENT){
			// see if the selected item is a directory and if 
			// so, descend to that directory.  
			if (SelectionIsDirectory()){
				SetDirNameFromSelection();
				setDirectoryWithUpdate(mDirName);
				repaint();
				return(true);
			}
		
			if (SelectionIsFile()){  	
				SetFileFromSelection();
				SetDirNameFromTextField();
				this.hide();
				mLoad = true;
				return(true);
			}
	}
	
	//----------------
	// Load
	//----------------
	if (target == mbtnLoad){
		// First we want to check whether a file name
		// has been entered in the text field.  If so, 
		// this will override any selection in the list.  
		if (TextFieldIsFile()){
			SetFileAndDirFromTextField();
			this.hide();
			mLoad=true;
			return(true);
		}
		
		// If the text field doesn't contain a file name, 
		// see if the selected item is a directory and if 
		// so, descend to that directory.  
		if (SelectionIsDirectory()){
			SetDirNameFromSelection();
			setDirectoryWithUpdate(mDirName);
			repaint();
			return(true);
		}

		// If selection is a file, return control to higher level, 
		// otherwise, do nothing.  
		if (SelectionIsFile()){  	
			SetFileFromSelection();
			SetDirNameFromTextField();
			this.hide();
			mLoad = true;
			return(true);
		}
		else{
			// Note, TextField will be a directory even as SelectionIsFile
			// so it is necessary that the file test occur first. 
			if (TextFieldIsDirectory()){
				// See if it is different from current directory
				if (mTextField.getText() != mDirName){
					SetDirNameFromTextField();
					setDirectoryWithUpdate(mDirName);
				}
			}
		}
	}
	
	//------------------
	// TextFieldUpdate
	//------------------
	if (target == mTextField){
		// First we want to check whether a file name
		// has been entered in the text field.  If so, 
		// this will override any selection in the list.  
		if (TextFieldIsFile()){
			SetFileAndDirFromTextField();
			this.hide();
			mLoad=true;
			return(true);
		}
		
		if (TextFieldIsDirectory()){
			// See if it is different from current directory
			if (mTextField.getText() != mDirName){
				SetDirNameFromTextField();
				setDirectoryWithUpdate(mDirName);
			}
		}
	}
	
	//-------------
	// Cancel
	//-------------
	if (target == mbtnCancel){
		this.hide();
		mLoad = false;
	}
	
	//-------------
	// Update
	//-------------
	if (target == mbtnUpdate){
		setDirectoryWithUpdate(mDirName);
		repaint();
	}
	
	return(true);
}

}//---| End FastFileDialog |----------------------------------------------------



