//================================================================================
//  SCFTrace.java
//--------------------------------------------------------------------------------
// Encapsulates the information from a SCF file, including methods to read the
// file. 
//
//--------------------------------------------------------------------------------
// SCFTrace and ABITrace are adapted from similar classes in:
// Trace Applet, by Eugen Buehler.   Release 1, 6/30/96 
//
// Adapted by K. James Durbin, Baylor College of Medicine.  7/17/97
//================================================================================
//
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//       1         2         3         4         5         6         7         8
//

import java.awt.*;
import java.io.*;
import Trace;
import SCFHeader;

//------------------------------------------------------------------------------
// * SCFTrace
//------------------------------------------------------------------------------
//
public class SCFTrace extends Trace{

	 SCFHeader	Header = new SCFHeader();
	 
	 private int		bytesRead;
	 private String 	comments, baseSequence;
	 private DataInputStream filterIN;
	 
	 public static final int MagicNum = 
		  (((((int)'.'<<8)+(int)'s'<<8)+(int)'c'<<8)+(int)'f');
	 
	 //------------------------------------------------------
	 // * SCFTrace (SCFTrace)
	 //------------------------------------------------------
	 //
	 public SCFTrace(InputStream in) throws IOException {
		  //System.out.println("SCFTrace about to read data..");
		  filterIN = new DataInputStream(in);
		  read();
		  //System.out.println("SCFTrace data read.");
	 }
	 	
	 //------------------------------------------------------
	 // * read (SCFTrace)
	 //------------------------------------------------------
	 //
	 public void read() throws IOException{
		  
		  bytesRead = 0;
		  
		  bytesRead = Header.read(filterIN);


		  // System.out.print("basesOffset="+Header.basesOffset+" SamplesOffset="+Header.SamplesOffset);
		  //System.out.println("commentsOffset="+Header.commentsOffset);
		  
		  // determine order of file components and then read them
		  if ((Header.basesOffset < Header.SamplesOffset) && 
				(Header.basesOffset < Header.commentsOffset)) {
				readBases();
				if (Header.SamplesOffset < Header.commentsOffset) {
					 readSample();
					 readComments();
				}
				else {
					 readComments();
					 readSample();
				}
		  }
		  else if (Header.SamplesOffset < Header.commentsOffset) {
				readSample();
				if (Header.basesOffset < Header.commentsOffset) {
					 readBases();
					 readComments();
				}
				else {
					 readComments();
					 readBases();
				}
		  }
		  else {
				readComments();
				if (Header.SamplesOffset < Header.basesOffset) {
					 readSample();
					 readBases();
				}
				else {
					 readBases();
					 readSample();
				}
		  }
	 }	
	 
	 //------------------------------------------------------
	 // * readBases
	 //------------------------------------------------------
	 //
	 private void readBases() throws IOException
	 {
		  int i;
		  filterIN.skipBytes(Header.basesOffset - bytesRead);
		  basePosition = new int[Header.baseNum];
		  base = new int[Header.baseNum];
		  
		  // SCF version < 3.0 has interleaved bases, 3.0 has grouped bases
		  System.out.println("SCF Version : "+Header.versionNum);
		  System.out.println("Number of bases: "+Header.baseNum); 
		  if (Header.versionNum < 3.0){
				for (i=0;i<Header.baseNum;i++) {
					 
					 basePosition[i] = filterIN.readInt();
					 filterIN.skipBytes(4);
					 					 
					 // We convert base letters {A,T,C,G,N} to an index {0,1,2,3,4} so that 
					 // it can be used directly to index a 2D array of trace values.  
					 base[i] =  baseNumFromLetter[filterIN.readUnsignedByte()];
					 filterIN.skipBytes(3); // Spare bytes. 
				}
		  }else{				
				for( i = 0; i < Header.baseNum;i++){
                basePosition[i] = filterIN.readInt();			  				
				}
				
				//filterIN.skipBytes( 4 * Header.baseNum );
				// Qualities
				quality = new int[Header.baseNum];
				for(i = 0;i< 4;i++){
				  for(int j = 0;j < Header.baseNum;j++){
					 int bq = filterIN.readUnsignedByte();
					 if (bq > 0){
						quality[j] = bq;
					 }
				  }
				}

				//System.out.println(">BaseQuality");
				//for(i = 0;i< Header.baseNum;i++){
				// System.out.print(quality[i]+" ");
				//  if ((i%25) == 0) System.out.println(" ");
				//	}
				            
            for( i = 0; i < Header.baseNum;i++){
                base[i] = (char) baseNumFromLetter[filterIN.readUnsignedByte()];
				}
            filterIN.skipBytes( 3 * Header.baseNum );				
		  }
		  baseSequence = String.valueOf( base );
        bytesRead = Header.basesOffset + Header.baseNum * 12;
	 }


	 //------------------------------------------------------
	 // * readSample
	 //------------------------------------------------------
	 private void readSample()
        throws IOException
    {
        int i = 0;

        filterIN.skipBytes( Header.SamplesOffset - bytesRead );

		  // Allocate space for four channels of trace data...
		  TraceData = new int[4][Header.sampleNum];

        if( Header.sampleSize == 2 && (double) Header.versionNum < 3.0 )
        {
            for( i = 0; i < Header.sampleNum; ++i )
            {
                TraceData[A][i] = filterIN.readUnsignedShort();
                TraceData[C][i] = filterIN.readUnsignedShort();
                TraceData[G][i] = filterIN.readUnsignedShort();
                TraceData[T][i] = filterIN.readUnsignedShort();
            }
        }
        else if( Header.sampleSize == 1 && (double) Header.versionNum < 3.0 )
        {
            for( i = 0; i < Header.sampleNum; ++i )
            {
                TraceData[A][i] = filterIN.readUnsignedByte();
                TraceData[C][i] = filterIN.readUnsignedByte();
                TraceData[G][i] = filterIN.readUnsignedByte();
                TraceData[T][i] = filterIN.readUnsignedByte();
            }
        }
        else if( Header.sampleSize == 2 && (double) Header.versionNum >= 3.0 )
        {
            for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[A][i] = filterIN.readShort();
            }
				TraceData[A] = deltaDelta( TraceData[A] );

            for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[C][i] = filterIN.readShort();
				}
            TraceData[C] = deltaDelta(TraceData[C]);

            for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[G][i] = filterIN.readShort();
				}
            TraceData[G] = deltaDelta(TraceData[G]);
            for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[T][i] = filterIN.readShort();
				}
            TraceData[T] = deltaDelta( TraceData[T] );
        }
        else if( Header.sampleSize == 1 && (double) Header.versionNum >= 3.0 ){
				for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[A][i] = filterIN.readByte();
				}
            TraceData[A] = deltaDelta(TraceData[A]);

            for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[C][i] = filterIN.readByte();
				}
            TraceData[C] = deltaDelta( TraceData[C] );
           
				for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[G][i] = filterIN.readByte();
				}
            TraceData[G] = deltaDelta(TraceData[G] );

            for( i = 0; i < Header.sampleNum; ++i ){
                TraceData[T][i] = filterIN.readByte();
				}
            TraceData[T] = deltaDelta( TraceData[T] );
        }
        bytesRead = Header.SamplesOffset + Header.sampleNum * Header.sampleSize * 4;
    }
	 	 
	 //------------------------------------------------------
	 // * readComments
	 //------------------------------------------------------
	 //
	 private void readComments() throws IOException {
		  int i;
		  char[] cArray = new char[Header.commentsSize];
		  
		  //System.out.println("reading comments.");
		  
		  filterIN.skipBytes(Header.commentsOffset - bytesRead);
		  for (i=0;i<Header.commentsSize;i++) {
				cArray[i] = (char) filterIN.readUnsignedByte();
		  }
		  comments = String.valueOf(cArray);
		  bytesRead = Header.commentsOffset + Header.commentsSize;
	 }
	 
	 
	 //------------------------------------------------------
	 // * Local Access Methods
	 //------------------------------------------------------
	 //
	 public int getcodeSet() {
		  return Header.codeSet;
	 }
	 
	 public String getComments() {
		  return comments;
	 }
	 
	 public String getVersion() {
		  return Header.versionString;
	 }


	 
	 private int[] deltaDelta(int[] DIn) {
		  int i = 0;
		  for (int j = 0; j < DIn.length - 1; j++) {
				DIn[j] = DIn[j] + i;
				i = DIn[j];
		  }
		  i = 0;
		  for (int k = 0; k < DIn.length - 1; k++) {
				DIn[k] = DIn[k] + i;
				i = DIn[k];
		  }
		  return DIn;
    }
    

	 
}//-------| End SCFTrace class |----------------------------------------------------
