//================================================================================
//  Trace.java
//--------------------------------------------------------------------------------
//  Abstract class encapsulating standardized chromatogram data structures
//  and access methods.  Specific file formats (or other non-file sources)  
//  will derive from this and add methods for parsing the specific format. 
//  
//--------------------------------------------------------------------------------
// This class is adapted from: Trace Applet, Release 1, 6/30/96
//                             by Eugen Buehler 
//
// Adapted by K. James Durbin, Baylor College of Medicine.  7/17/97
//================================================================================
//
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//       1         2         3         4         5         6         7         8

import java.io.*;
import java.awt.*;

public abstract class Trace {
	protected int A=0,C=1,G=2,T=3,N=4;
	protected int TraceData[][];
	protected int basePosition[];
	protected int base[];
	protected int quality[];
		
	// When read in from a file we convert base letters to base numbers {0..4}
	// This gains efficiency in many places.  The ASCII code for A,C,G,T,N all
	// fall in the range 65-84, so an array of 100 elts is large enough to 
	// index them without subtracting an offset (normally I wouldn't be so
	// nit pickingly efficiency conscious, but drawing is very slow as it 
	// is that anything that might help is desired. 
	
	// 11/18/97  It is far from obvious to me now where the supposed 
	// efficiency of using 0...4 for base numbers is supposed to come 
	// from.  In fact, the baseColor array is the closest thing to 
	// something that is indexed on the base number, and we could as 
	// well allocate an array of 100 colors to store this as we can 
	// go through this letter from num conversion every time we 
	// want to draw a base. 
	protected int  baseNumFromLetter[] = new int[100];

	// To display bases these numbers need to be converted back to letters 
	// as follows
	protected char baseLetterFromNum[]={'A','C','G','T','N'};
	
	public abstract void read() throws IOException;
	public abstract String getVersion();
	public abstract String getComments();
	
//------------------------------------------------------
// Trace (Trace)
//------------------------------------------------------
// 
public Trace(){

	// Initialize the baseNum hash.  
	for (int i = 0;i<baseNumFromLetter.length;i++){
		baseNumFromLetter[i] = 4;  // N and others...
	}	
	// Create a simple hash 
	baseNumFromLetter[65]= 0;  	// A
	baseNumFromLetter[67]= 1;	// C
	baseNumFromLetter[71]= 2;	// G
	baseNumFromLetter[84]= 3;	// T
	
}

//------------------------------------------------------
// * Access Methods
//------------------------------------------------------
//
public int getTraceData(int base,int Index){
	return(TraceData[base][Index]);
}

public int getBase(int Index) {
	return (base[Index]);
}

public int getQual(int Index){
	return(quality[Index]);
}

public int getNumBases() {
	return ((int) base.length);
}

public int getTraceLength() {
	return ((int)TraceData[A].length);
	//return Header.sampleNum;
}

public int getBasePosition(int Index) {
	return(basePosition[Index]);
}

	 //-----------------------------
	 // * CopyBogusQuality
	 //-----------------------------
	 // When no quality is given, just init everythign to zero. 
	 // 
	 public void CopyBogusQuality()
	 {
		  quality = new int[base.length];		  
	 }


	 //-------------------------------
	 // * CopyQuality
	 //-------------------------------
	 public void CopyQuality(int QualitySource[],int NumValues)
	 {
		  // If the bases have been initialized, do a sanity check on the 
		  // number of quality values.  
		  if (base.length != 0){
				if (NumValues != base.length){
					 System.err.println("Warning!  Number of bases does not match quality.");
					 System.err.println("NumBases="+base.length+" NumQual="+NumValues);
					 
					 for(int i = 0;i< NumValues;i++){
						  System.err.print(QualitySource[i]+" ");
						  if ((i%25)==0){
								System.err.println();
						  }
					 }
				}
		  }
		  quality = new int[NumValues];
		  System.arraycopy(QualitySource,0,quality,0,NumValues);
	 }

//------------------------------------------------------
// * GetLastBaseBeforeSample
//------------------------------------------------------
// Base information such as sample position where base 
// was called, is stored in a simple array. Since this 
// array is sorted by position, binary search provides a 
// reasonaly efficient way to go from sample position 
// to base. A hand coded hash might be faster, but the 
// Java hash is definitely too slow for this purpose. 
// Using binary search is about 10x faster than linear 
// search on the typical trace which is a noticable 
// improvement.  
//
public int GetLastBaseBeforeSample(int sampleNum){
	int first,last,mid;
	
	first = 1;
	last = base.length-1;
	//System.out.println("Base length="+base.length);
	mid = (last-first)/2;
	
	// Binary search until either land exactly on sample
	// or narrow down to two possible bases...
	while ((last-first)>=2){
		if (sampleNum < basePosition[mid]){
			last = mid;
		} else if (sampleNum > basePosition[mid]){
			first = mid;
		}
		else{
			// we nailed it.  
			return(mid);
		}
		mid = first+(last-first)/2;
	}
	
	// Narrowed down to two bases, now choose one...
	if ((sampleNum > basePosition[first])&&
		 (sampleNum < basePosition[last])){
		return(first);
	}
	if (sampleNum > basePosition[last]){
		return(last);
	}
	
	// (Sanity check for testing...)
	//if (sampleNum  < basePosition[first]){
	//	System.out.println("Error in binary search!");
	//	return(0);
	//}
	
	return(0);
}	
	
//-----------------------------------------------------
// * GetNextBaseAfterSample
//-----------------------------------------------------
// See GetLastBaseBeforeSample.  
// 
public int GetNextBaseAfterSample(int sampleNum){
	int first,last,mid,bugCount=0;
	first = 1;
	last = base.length-1;
	mid = (last-first)/2;
	
	// Binary search until hit it exactly or narrow down
	// to two candidates...
	while ((last-first)>=2){
		if (sampleNum < basePosition[mid]){
			last = mid;
		} else if (sampleNum > basePosition[mid]){
			first = mid;
		}
		else{
			// hit it exactly ...  
			return(mid);
		}
		mid = first+(last-first)/2;
	}
	
	// Narrowed down to two candidates, choose one...
	if (sampleNum  < basePosition[first]){
		return(first);
	}
	if ((sampleNum > basePosition[first])&&
		 (sampleNum < basePosition[last])){
		 return(last);
	}
	
	// (Sanity check for debugging...)
	//if (sampleNum > basePosition[last]){
	//	System.out.println("Error in binary search!");
	//	return(0);
	//}
	return(0);
}

} // end class. 
